#############################################################################
# menubuttons.tcl Copyright 2004 r0lZ
# Menu VOB buttons infos.
#
# This file is part of PgcEdit.
#
# PgcEdit is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# PgcEdit is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#############################################################################

## For each VTS, the following variables are set:
## ::buttons::butinfo($vts,errorcode)	;# 0=OK
## ::buttons::butinfo($vts,errormsg)
## ::buttons::butinfo($vts,numvobids)
## For each VOB ID
## ::buttons::butinfo($vts,$vobid,numcellids)
## For each Cell ID
## ::buttons::butinfo($vts,$vobid,$cellid,navpack)	;# first NAV pack number for current VOB/Cell ID
## ::buttons::butinfo($vts,$vobid,$cellid,numbuts)	;# number of buttons (in first group)
## ::buttons::butinfo($vts,$vobid,$cellid,numgroups)
## ::buttons::butinfo($vts,$vobid,$cellid,startbutn)
## ::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn)
## ::buttons::butinfo($vts,$vobid,$cellid,forceactionbutn)
## For eacg group
## ::buttons::butinfo($vts,$vobid,$cellid,$group,grouptype)	;# 1=normal, 2=letterbox, 4=pan&scan (ored together)
## For each button
## ::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,autoaction) ;# 0 or 1
## ::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,coords)	;# Button position: {x_start y_start - x_end y_end}
## ::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,udlr)		;# Adjacent buttons: {up down left right}
## ::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,command)	;# Button command: {XX XX XX XX XX XX XX XX}
## and
## ::pgcs($vts,$pgc,cellpostable)		;# Cell Position Table, in binary

namespace eval ::menubuttons {


#############################################################################
## Procedure:  dump_pgc_menu_buttons
proc dump_pgc_menu_buttons {vts numcells cellpostable indentstring} {
	if {$numcells == 0} {return ""}
	if {$::tcl_platform(platform) != "windows"} {return ""}

	set txt "${indentstring}********** menu buttons commands:\n"
	for {set cell 0} {$cell < $numcells} {incr cell} {
		set idx [expr $cell * 4]
		set vobid [get_int cellpostable $idx 2]
		incr idx 3
		set cellid [get_int cellpostable $idx 1]
		if [info exists ::buttons::butinfo($vts,$vobid,$cellid,numbuts)] {
			set numbuts $::buttons::butinfo($vts,$vobid,$cellid,numbuts)
			if {$numbuts == 0} {continue}
			set txt "${txt}${indentstring}VOB ID $vobid,  Cell ID $cellid  (First NAV pack=$::buttons::butinfo($vts,$vobid,$cellid,navpack))\n"

			set numgroups $::buttons::butinfo($vts,$vobid,$cellid,numgroups)
			set txt "${txt}${indentstring}$numgroups Group"
			if {$numgroups > 1} {
				set txt "${txt}s"
			}
			set txt "$txt,  Start button=$::buttons::butinfo($vts,$vobid,$cellid,startbutn),"
			set txt "$txt  Select button=$::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn),"
			set txt "$txt  Action button=$::buttons::butinfo($vts,$vobid,$cellid,forceactionbutn)\n"
			for {set butnum 1} {$butnum <= $numbuts} {incr butnum} {
				for {set group 1} {$group <= $numgroups} {incr group} {
					set cmdtxt "${indentstring}button [format "%2d" $butnum]: "
					set hex $::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,command)
					set cmdtxt "${cmdtxt}\[$hex\]  [::disasm::disasm $hex] "
					if $::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,autoaction) {
						set cmdtxt "$cmdtxt*** Autoaction "
					} else {
						set cmdtxt "$cmdtxt"
					}
					set cmdtext($group) $cmdtxt
				}
				set samecmds true
				if {$numgroups > 1} {
					if {$cmdtext(1) != $cmdtext(2)} {set samecmds false}
					if {$numgroups > 2} {
						if {$cmdtext(1) != $cmdtext(3) || $cmdtext(2) != $cmdtext(3)} {set samecmds false}
					}
				}
				if {! $samecmds} {
					for {set group 1} {$group <= $numgroups} {incr group} {
						set txt "${txt}$cmdtext($group)*** Group $group ([grouptype_descr $::buttons::butinfo($vts,$vobid,$cellid,$group,grouptype)])\n"
					}
				} else {
					set txt "${txt}$cmdtext(1)\n"
				}
			}
		}
	}
	return $txt
}

#############################################################################
## Procedure:  dump_all_button_infos
proc dump_all_button_infos {header} {
	set out "$header"
	set out "$out\nDVD-TEXT General Name: \"[::utils::get_General_Name]\""
	set out "$out\nProvider ID: \"[::utils::get_provider_ID]\""
	set out "$out\nNumber of VTS: $::pgcs(numvts)\n"

	# VIDEO_TS.IFO
	show_msg "Dumping VMGM buttons..."
	if {$::buttons::butinfo(0,errorcode) == 0} {
		set out "$out\nVMG (VIDEO_TS.VOB)"
		set out "$out\n------------------"
		set out "$out\n[dump_vts_button_infos 0]"
	}

	# VTS_xx_0.IFO
	for {set vts 1} {$vts <= $::pgcs(numvts)} {incr vts} {
		show_msg "Dumping VTS $vts buttons..."
		if {$::buttons::butinfo($vts,errorcode) == 0} {
			set out "$out\nVTS $vts (VTS_[format %02i $vts]_0.VOB)"
			set out "$out\n--------------------"
			set out "$out\n[dump_vts_button_infos $vts]"
		}
	}
	show_msg ""
	return [string trimright $out "\n"]\n
}

#############################################################################
## Procedure:  dump_all_button_infos
proc dump_vts_button_infos {vts} {
	if {$vts == 0} {set head "VMGM"} else {set head "VTSM $vts"}
	set txt ""
	for {set vobid 1} {$vobid <= $::buttons::butinfo($vts,numvobids)} {incr vobid} {
		for {set cellid 1} {$cellid <= $::buttons::butinfo($vts,$vobid,numcellids)} {incr cellid} {
			set txt "$txt  $head,  VOB ID $vobid,  Cell ID $cellid  (First NAV pack=$::buttons::butinfo($vts,$vobid,$cellid,navpack))\n"
			if {$::buttons::butinfo($vts,$vobid,$cellid,numbuts) == 0} {
				set txt "$txt  No buttons.\n"
			} else {
				set txt "$txt  $::buttons::butinfo($vts,$vobid,$cellid,numgroups) Group"
				if {$::buttons::butinfo($vts,$vobid,$cellid,numgroups) > 1} {
					set txt "${txt}s"
				}
				set txt "$txt,  Start but=$::buttons::butinfo($vts,$vobid,$cellid,startbutn),"
				set txt "$txt  Select but=$::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn),"
				set txt "$txt  Action but=$::buttons::butinfo($vts,$vobid,$cellid,forceactionbutn)\n"
				set numgroups $::buttons::butinfo($vts,$vobid,$cellid,numgroups)
				for {set butnum 1} {$butnum <= $::buttons::butinfo($vts,$vobid,$cellid,numbuts)} {incr butnum} {
					for {set group 1} {$group <= $numgroups} {incr group} {
						set cmdtxt "    button [format "%2d" $butnum]: "
						set hex $::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,command)
						set cmdtxt "${cmdtxt}\[$hex\]  [::disasm::disasm $hex] "
						if $::buttons::butinfo($vts,$vobid,$cellid,$group,$butnum,autoaction) {
							set cmdtxt "$cmdtxt*** Autoaction "
						} else {
							set cmdtxt "$cmdtxt"
						}
						set cmdtext($group) $cmdtxt
					}
					set samecmds true
					if {$numgroups > 1} {
						if {$cmdtext(1) != $cmdtext(2)} {set samecmds false}
						if {$numgroups > 2} {
							if {$cmdtext(1) != $cmdtext(3) || $cmdtext(2) != $cmdtext(3)} {set samecmds false}
						}
					}
					if {! $samecmds} {
						for {set group 1} {$group <= $numgroups} {incr group} {
							set txt "${txt}$cmdtext($group)*** Group $group ([grouptype_descr $::buttons::butinfo($vts,$vobid,$cellid,$group,grouptype)])\n"
						}
					} else {
						set txt "${txt}$cmdtext(1)\n"
					}
				}
			}
			set txt "$txt\n"
		}
	}
	set txt "$txt\n"
}

proc show_menu_buttons_info {} {
	if {! [info exists ::dvddir]} {bell; return}
	::main::showtext .buttoncmds "Menu buttons commands" [dump_all_button_infos "Menu buttons commands"]
}

proc grouptype_descr {type} {
	set descr ""
	if {[expr $type & 1] == 1} {set descr "${descr}+normal"}
	if {[expr $type & 2] == 2} {set descr "${descr}+letterbox"}
	if {[expr $type & 4] == 4} {set descr "${descr}+pan&scan"}
	return [string trim $descr "+"]
}

###################################################
## Display Buttons GUI
###################################################

proc display_current_menu_buttons {} {
	if {! [info exists ::dvddir]} {
		bell
		return
	}
	if {$::scratchbook_active} {
		bell
		return 0
	}
	if {$::tcl_platform(platform) != "windows"} {
		show_msg "Menu buttons: Available only under Windows"
		bell
		return 0
	}
	if {[lindex $::editor_cmds_from 0] != "menu"} {
		show_msg "Menu buttons: Available only for Menu PGCs"
		bell
		return 0
	}
	set indexlist [split [lindex $::editor_cmds_from 1] ","]
	
	set ::buttons::groupnum 1

	display_menu_buttons [lindex $indexlist 0] [lindex $indexlist 1] [lindex $indexlist 2] 1024 "info" 0
}

proc display_menu_buttons {vts lu pgc sprm8 mode show_cell} {
	set utiltitle "Menu buttons"
	set w .menubuttons

	if {! [info exists ::buttons::groupnum]} {set ::buttons::groupnum 1}

	if {$::menus($vts,$lu,$pgc,numcells) == 0} {
		show_msg "Menu buttons: This menu has no video cells"
		bell
		return 0
	}
	if {$::menus($vts,$lu,$pgc,cellpostable) == ""} {
		show_msg "Menu buttons: This menu has no video cell position table"
		bell
		return 0
	}
	if {$::buttons::butinfo($vts,errorcode) != 0} {
		show_msg "Menu buttons: Error when parsing Menu VOB file: $::buttons::butinfo($vts,errormsg)"
		bell
		return 0
	}

	set programmap $::menus($vts,$lu,$pgc,programmap)
	for {set p 1} {$p <= $::menus($vts,$lu,$pgc,numprograms)} {incr p} {
		set cell_program([get_int programmap [expr $p-1] 1]) $p
	}
	set old_pg 1
	for {set c 1} {$c <= $::menus($vts,$lu,$pgc,numcells)} {incr c} {
		if [info exists cell_program($c)] {
			set old_pg $cell_program($c)
		} else {
			set cell_program($c) $old_pg
		}
	}

	set buttonsfound false
	set cellpostable $::menus($vts,$lu,$pgc,cellpostable)
	if {$mode == "info"} {
		set menulabels {}
		set menucommands {}
		set ::menubuttons::menutargets {}
		set cell 0
		for {set c 1} {$c <= $::menus($vts,$lu,$pgc,numcells)} {incr c} {
			set idx [expr ($c-1) * 4]
			set this_vobid [get_int cellpostable $idx 2]
			incr idx 3
			set this_cellid [get_int cellpostable $idx 1]
			if [info exists ::buttons::butinfo($vts,$this_vobid,$this_cellid,numbuts)] {
				set this_numbuts $::buttons::butinfo($vts,$this_vobid,$this_cellid,numbuts)
				if {$this_numbuts == 0} {continue}
				set buttonsfound true
				if {$cell == 0} {
					set cell $c
					set vobid $this_vobid
					set cellid $this_cellid
					set numbuts $this_numbuts
				}
				lappend menulabels "Program $cell_program($c), Cell $c:  VOB ID, $this_vobid, Cell ID $this_cellid ($this_numbuts buttons)"
				lappend menucommands "::menubuttons::display_cell $w $vts $lu $pgc $c $cell_program($c) $this_vobid $this_cellid 1024"
				lappend ::menubuttons::menutargets "$cell_program($c) $c $this_vobid $this_cellid"
			}
		}
		if {! $buttonsfound} {
			show_msg "Menu buttons: This menu has no buttons"
			bell
			return 0
		}
	} else {
		set cell $show_cell
		set idx [expr ($cell-1) * 4]
		set vobid [get_int cellpostable $idx 2]
		incr idx 3
		set cellid [get_int cellpostable $idx 1]
		if [info exists ::buttons::butinfo($vts,$vobid,$cellid,numbuts)] {
			set numbuts $::buttons::butinfo($vts,$vobid,$cellid,numbuts)
			if {$numbuts == 0} {
				show_msg "Menu buttons: This menu has no buttons"
				bell
				return 0
			}
		}
	}

	### GUI
	button .b
	set font [.b cget -font]
	destroy .b
	set ::menubuttons::minxsize [font measure $font "00"]
	set ::menubuttons::minysize [expr [font metrics $font -ascent] - [font metrics $font -descent]]
	
	set nullimg {
		#define null_width 1
		#define null_height 1
		static unsigned char null_bits[] = {
			0x00};
	}
	image create bitmap ::menubuttons::bitmapnull -background {} -data $nullimg

	if [winfo exists $w] {destroy $w}
	toplevel $w -class Dialog
	wm withdraw $w
	wm title $w $utiltitle
	wm transient $w .
	if {$::tcl_platform(platform) == "windows"} {
		catch {wm attributes $w -toolwindow 1}
	}
	wm protocol $w WM_DELETE_WINDOW {set ::menubuttons::menubuttons_done 1}
	bind $w <Escape> {set ::menubuttons::menubuttons_done 1}
	if {$mode == "trace"} {
		bind $w <Return> {set ::menubuttons::menubuttons_done 1}
	}

	if {$mode == "trace"} {
		label $w.this -bd 1 -relief sunken -height 1 -textvariable ::menubuttons::lab1_var
	} else {
		menubutton $w.this -menu $w.this.select_cell -textvariable ::menubuttons::lab1_var -pady 1
		menu $w.this.select_cell
		for {set i 0} {$i < [llength $menulabels]} {incr i} {
			$w.this.select_cell add radiobutton -label [lindex $menulabels $i] \
					-command [lindex $menucommands $i] \
					-variable ::menubuttons::cell_select_menu_item -value [expr $i+1]
		}
		set ::menubuttons::cell_select_menu_item 1
	}
	pack  $w.this -fill x -padx 4 -pady 2 -ipady 2

	label $w.labinfo -bd 1 -relief sunken -height 1 \
		-textvariable ::menubuttons::lab2_var
	pack  $w.labinfo -fill x -padx 4 -pady 2 -ipady 2

	set scrndim [get_screen_dims $vts]
	pack [frame $w.video -bd 2 -relief ridge] -padx 4 -pady 4
	pack [frame $w.video.f -width [expr round([lindex $scrndim 0] / 3.0 * 2)] \
			-height [expr round([lindex $scrndim 1] / 3.0 * 2)] -background gray]

	frame $w.f2 -bd 2 -relief groove
	pack  $w.f2 -padx 4 -pady 2 -fill x -expand 1
	label $w.f2.labcmd -bd 1 -relief sunken -height 3 -textvariable ::menubuttons::command
	pack  $w.f2.labcmd -fill x -padx 4 -pady 2
	frame $w.f2.f
	pack  $w.f2.f -fill x -expand 1
	checkbutton $w.f2.f.auto -state disabled -text "Auto Action" -disabledforeground black \
			-variable ::menubuttons::autoaction
	pack   $w.f2.f.auto -side left
	if {$mode == "info"} {
		button $w.f2.f.follow -text "  Jump to Target  " -pady 0	;# -command set later
		pack   $w.f2.f.follow -side right -padx 4 -pady 2
	}

	# buttons
	frame  $w.fb
	pack   $w.fb -padx 2 -pady 4 -fill x -expand 1
	label  $w.fb.lab3 -text "Group" -padx 0
	pack   $w.fb.lab3 -side left
	radiobutton $w.fb.g1 -variable ::buttons::groupnum -value 1 -text "1" -command \
			"::menubuttons::display_cell $w $vts $lu $pgc $cell_program($cell) $vobid $cellid 1024"
	pack   $w.fb.g1 -side left
	radiobutton $w.fb.g2 -variable ::buttons::groupnum -value 2 -text "2" -command \
			"::menubuttons::display_cell $w $vts $lu $pgc $cell_program($cell) $vobid $cellid 1024"
	pack   $w.fb.g2 -side left
	radiobutton $w.fb.g3 -variable ::buttons::groupnum -value 3 -text "3" -command \
			"::menubuttons::display_cell $w $vts $lu $pgc $cell_program($cell) $vobid $cellid 1024"
	pack   $w.fb.g3 -side left	
	label  $w.fb.lab1 -text "   Button" -padx 0
	pack   $w.fb.lab1 -side left
	label  $w.fb.lab2 -textvariable ::buttons::default_button -padx 0 -width 2
	pack   $w.fb.lab2 -side left
	button $w.fb.prev -text "<" -width 2 -pady 0 \
		-command "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid \[expr \$::buttons::default_button - 1\] true"
	pack   $w.fb.prev -side left -padx 0
	button $w.fb.next -text ">" -width 2 -pady 0 \
		-command "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid \[expr \$::buttons::default_button + 1\] true"
	pack   $w.fb.next -side left -padx 0
	frame  $w.fb.pad1
	pack   $w.fb.pad1 -side left -fill x -expand 1
	button $w.fb.ok -text "OK" -width 13 -command {set ::menubuttons::menubuttons_done 1}
	pack   $w.fb.ok -side left -padx 4

	update
	wm minsize $w [winfo reqwidth $w] [winfo reqheight $w]
	wm maxsize $w [expr [winfo reqwidth $w] + 300] [winfo reqheight $w]
	if {! [info exists ::menubuttons::win_geometry]} {
		wm geometry $w $::config(menubuttons_win_geometry)
	} else {
		wm geometry $w $::menubuttons::win_geometry
	}
	
	display_cell $w $vts $lu $pgc $cell $cell_program($cell) $vobid $cellid 1024
	
	wm deiconify $w
	grab set $w

	tkwait variable ::menubuttons::menubuttons_done
	$w.video.f.b$::buttons::default_button configure -background white -relief sunken
	if {$mode == "trace"} {
		after 500 "set ::menubuttons::menubuttons_done 2"
		tkwait variable ::menubuttons::menubuttons_done
	}
	set ::menubuttons::win_geometry [wm geometry $w]
	destroy $w
	return $::buttons::default_button
}

proc display_cell {w vts lu pgc cell program vobid cellid sprm8} {
	set numbuts $::buttons::butinfo($vts,$vobid,$cellid,numbuts)
	if {$vts == 0} {set vtslabel "VMGM"} else {set vtslabel "VTSM $vts"}
	set ::menubuttons::lab1_var "$vtslabel / LU $lu / PGC $pgc,  PG $program,  Cell $cell :  VOB ID $vobid, Cell ID $cellid (first NAV pack $::buttons::butinfo($vts,$vobid,$cellid,navpack))"
	set ::menubuttons::lab2_var "Group $::buttons::groupnum ([grouptype_descr $::buttons::butinfo($vts,$vobid,$cellid,$::buttons::groupnum,grouptype)]), $numbuts buttons, force select button $::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn), force action button $::buttons::butinfo($vts,$vobid,$cellid,forceactionbutn)"

	if {$::buttons::butinfo($vts,$vobid,$cellid,numgroups) < 3} {
		$w.fb.g3 configure -state disabled
		if {$::buttons::groupnum == 3} {set ::buttons::groupnum 2}
	} else {
		$w.fb.g3 configure -state normal -command \
			"::menubuttons::display_cell $w $vts $lu $pgc $cell $program $vobid $cellid 1024"
	}
	if {$::buttons::butinfo($vts,$vobid,$cellid,numgroups) < 2} {
		$w.fb.g2 configure -state disabled
		if {$::buttons::groupnum == 2} {set ::buttons::groupnum 1}
	} else {
		$w.fb.g2 configure -state normal -command \
			"::menubuttons::display_cell $w $vts $lu $pgc $cell $program $vobid $cellid 1024"
	}
	$w.fb.g1 configure -command "::menubuttons::display_cell $w $vts $lu $pgc $cell $program $vobid $cellid 1024"
	$w.fb.prev configure -command \
		"::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid \[expr \$::buttons::default_button - 1\] true"
	$w.fb.next configure -command \
		"::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid \[expr \$::buttons::default_button + 1\] true"
	$w.f2.f.follow configure -command \
		"::menubuttons::jump_to_target_inside_PGC $w $vts $lu $pgc $cell $program \$::menubuttons::command"
			
	set scrndim [get_screen_dims $vts]
	destroy $w.video.f
	pack [frame $w.video.f -width [expr round([lindex $scrndim 0] / 3.0 * 2)] \
			-height [expr round([lindex $scrndim 1] / 3.0 * 2)] -background gray]

	for {set but $numbuts} {$but > 0} {incr but -1} {
		display_button $w $vts $lu $pgc $vobid $cellid $but false
	}
	if {$::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn) != 0} {
		set ::buttons::default_button $::buttons::butinfo($vts,$vobid,$cellid,forceselectbutn)
	} elseif {$sprm8 >= 1024} {
		set ::buttons::default_button [expr $sprm8 / 1024]
	} else {
		set ::buttons::default_button 1
	}
	display_button $w $vts $lu $pgc $vobid $cellid $::buttons::default_button true
}

proc display_button {w vts lu pgc vobid cellid but default} {
	set group $::buttons::groupnum
	if {$but > 0 && $but <= $::buttons::butinfo($vts,$vobid,$cellid,numbuts)} {
		set coords $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,coords)
		set x_start [expr round([lindex $coords 0] / 3.0 * 2)]
		set y_start [expr round([lindex $coords 1] / 3.0 * 2)]
		set x_end   [expr round([lindex $coords 3] / 3.0 * 2)]
		set y_end   [expr round([lindex $coords 4] / 3.0 * 2)]
		set x_size  [expr $x_end - $x_start - 4]
		set y_size  [expr $y_end - $y_start - 4]
		update
		if  {$x_size < $::menubuttons::minxsize} {
			set x_size $::menubuttons::minxsize
			set x_end [expr $x_start + $::menubuttons::minxsize + 4]
			if {$x_end > [winfo width $w.video.f]} {
				incr x_start [expr [winfo width $w.video.f] - $x_end]
			}
		}
		if  {$y_size < $::menubuttons::minysize} {
			set y_size $::menubuttons::minysize
			set y_end [expr $y_start + $::menubuttons::minysize + 4]
			if {$y_end > [winfo height $w.video.f]} {
				incr y_start [expr [winfo height $w.video.f] - $y_end]
			}
		}
	
		if [winfo exists $w.video.f.b$but] {destroy $w.video.f.b$but}
		button $w.video.f.b$but -image ::menubuttons::bitmapnull -bd 2 -padx 0 -pady 0 \
				-compound center -highlightthickness 0 -overrelief sunken -text $but \
				-width $x_size -height $y_size -command "::menubuttons::display_button \
					$w $vts $lu $pgc $vobid $cellid $but true"
		if $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,autoaction) {
			$w.video.f.b$but configure -relief ridge
		}
		if $default {
			set ::buttons::default_button $but
			for {set i 1} {$i <= $::buttons::butinfo($vts,$vobid,$cellid,numbuts)} {incr i} {
				$w.video.f.b$i configure -background [$w.video cget -background]
			}
			if $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,autoaction) {
				$w.video.f.b$but configure -background lightgreen
			} else {
				$w.video.f.b$but configure -background lightblue
			}
			set ::menubuttons::autoaction $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,autoaction)
			set ::menubuttons::command "\[$::buttons::butinfo($vts,$vobid,$cellid,$group,$but,command)\]\n[::disasm::disasm $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,command)]"
			set udlr $::buttons::butinfo($vts,$vobid,$cellid,$group,$but,udlr)
			bind $w <Up>    "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid [lindex $udlr 0] true"
			bind $w <Down>  "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid [lindex $udlr 1] true"
			bind $w <Left>  "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid [lindex $udlr 2] true"
			bind $w <Right> "::menubuttons::display_button $w $vts $lu $pgc $vobid $cellid [lindex $udlr 3] true"

			::main::highlight_jump_cmd menu $vts,$lu,$pgc $::menubuttons::command 0 false
		}
		place $w.video.f.b$but -x $x_start -y $y_start
	}
}

proc get_screen_dims {vtsnum} {
	if {$vtsnum == 0} {
		set mat_table $::raw_sectors(0,VMGM_MAT)
	} else {
		set mat_table $::raw_sectors($vtsnum,VTSI_MAT)
	}
	set vattr [get_int mat_table 0x100 2]
	if {[expr $vattr & 0X1000] == 0} {
		set h 480
	} else {
		set h 576
	}
	#switch [expr ($vattr & 0X000C) / 4] {
	#	0 {set w 720}
	#	1 {set w 704}
	#	2 {set w 352}
	#	3 {set w 352; set h [expr $h / 2]}
	#}
	if {[expr ($vattr & 0X000C) / 4] == 0} {
		set w 720
	} else {
		set w 704
	}
	return "$w $h"
}

proc jump_to_target_inside_PGC {w vts lu pgc cell pg cmd} {
	set cmd [string range $cmd 1 23]
	set jumpfound 0
	set targetpg   -1
	set targetcell -1
	set cmdchk [lindex $cmd 0][string range [lindex $cmd 1] 1 1]
	switch $cmdchk {
		"201" {
			# Link*PG and Link*Cell
			set cmdchk2 [lindex $cmd 7]
			switch $cmdchk2 {
				"01" {
					# LinkTopCell
					set targetcell $cell
				}
				"02" {
					# LinkNextCell
					set targetcell [expr $cell + 1]
				}
				"03" {
					# LinkPrevCell
					set targetcell [expr $cell - 1]
				}
				"05" {
					# LinkTopPG
					set targetpg $pg
				}
				"06" {
					# LinkNextPG
					set targetpg [expr $pg + 1]
				}
				"07" {
					# LinkPrevPG
					set targetpg [expr $pg - 1]
				}			
			}
		}
		"206" {
			# LinkPGN
			set targetpg [expr 0x[lindex $cmd 7]]
		}
		"207" {
			# LinkCN
			set targetcell [expr 0x[lindex $cmd 7]]
		}
	}
	if {$targetpg != -1} {
		set hl_btnn [expr (0x[lindex $cmd 6][lindex $cmd 7] / 1024) * 1024]
		for {set i 0} {$i < [llength $::menubuttons::menutargets]} {incr i} {
			set entry [lindex $::menubuttons::menutargets $i]
			if {$targetpg == [lindex $entry 0]} {
				set jumpfound 1
				set ::menubuttons::cell_select_menu_item $i
				display_cell $w $vts $lu $pgc [lindex $entry 1] $targetpg [lindex $entry 2] [lindex $entry 3] $hl_btnn
				break
			}
		}
		if {! $jumpfound} bell
	}
	if {$targetcell != -1} {
		set hl_btnn [expr (0x[lindex $cmd 6][lindex $cmd 7] / 1024) * 1024]
		for {set i 0} {$i < [llength $::menubuttons::menutargets]} {incr i} {
			set entry [lindex $::menubuttons::menutargets $i]
			if {$targetpg == [lindex $entry 1]} {
				set jumpfound 1
				set ::menubuttons::cell_select_menu_item $i
				display_cell $w $vts $lu $pgc $targetcell [lindex $entry 0] [lindex $entry 2] [lindex $entry 3] $hl_btnn
				break
			}
		}
		if {! $jumpfound} bell
	}

	if {! $jumpfound} {
		set jumpfound [::main::highlight_jump_cmd menu $vts,$lu,$pgc \[$cmd\] 0 true]
	}
	return $jumpfound
}

}
